import sys #Print errors (sys.stderr.write)
import re #Regular expressions
import os #File and path checking

#Default settings - See Readme.txt and PictureViewer.cfg
Settings={'Include':'', 'Exclude':'', 'Extensions':'jpg;jpeg;png;bmp;gif;tiff', 'Sort':'Random', 'StartOn':'0' , 'SaveStartOn':'Yes', 'Position':'5', 'Background':'No', 'BGColor':'0', 'IsOpaque':'No', 'Translucency':'50', 'FitToScreen':'No', 'ShrinkToFit':'Yes', 'Interval':'10000', 'Cycle':'Yes', 'DisplayName':'No', 'NamePosition':'2', 'FGColor':'FFFFFF', 'TextUsesBG':'Yes', 'ExitOnActivity':'No', 'Preload':'Yes', 'LKeyQuit':'ESC', 'LKeyNameToCB':'C', 'LKeyInfo':'I', 'LKeyNext':'SPACE;RIGHT', 'LKeyPrevious':'LEFT', 'LKeyZoomIn':'=', 'LKeyZoomOut':'-', 'GKeyGetFocus':'CTRL+SHIFT+P', 'LKeyCycle':'ENTER', 'LKeyTranslucencyU':'NUM+', 'LKeyTranslucencyD':'NUM-', 'LKeyDisplayName':'N', 'LKeyPosition1':'NUM1', 'LKeyPosition2':'NUM2', 'LKeyPosition3':'NUM3', 'LKeyPosition4':'NUM4', 'LKeyPosition5':'NUM5', 'LKeyPosition6':'NUM6', 'LKeyPosition7':'NUM7', 'LKeyPosition8':'NUM8', 'LKeyPosition9':'NUM9', 'LKeyBackground':'B', 'LKeyIsOpaque':'O', 'LKeyFitToScreen':'F', 'LKeyShrinkToFit':'S', 'LKeyIntervalU':'[', 'LKeyIntervalD':']', 'LKeyNamePosition1':'CTRL+NUM1', 'LKeyNamePosition2':'CTRL+NUM2', 'LKeyNamePosition3':'CTRL+NUM3', 'LKeyNamePosition4':'CTRL+NUM4', 'LKeyNamePosition5':'CTRL+NUM5', 'LKeyNamePosition6':'CTRL+NUM6', 'LKeyNamePosition7':'CTRL+NUM7', 'LKeyNamePosition8':'CTRL+NUM8', 'LKeyNamePosition9':'CTRL+NUM9', 'LKeyTextUsesBG':'T', 'LKeyPreload':'P', 'GKeyQuit':'', 'GKeyNameToCB':'', 'GKeyInfo':'', 'GKeyNext':'', 'GKeyPrevious':'', 'GKeyZoomIn':'', 'GKeyZoomOut':'', 'GKeyCycle':'', 'GKeyTranslucencyU':'', 'GKeyTranslucencyD':'', 'GKeyDisplayName':'', 'GKeyPosition1':'', 'GKeyPosition2':'', 'GKeyPosition3':'', 'GKeyPosition4':'', 'GKeyPosition5':'', 'GKeyPosition6':'', 'GKeyPosition7':'', 'GKeyPosition8':'', 'GKeyPosition9':'', 'GKeyBackground':'', 'GKeyIsOpaque':'', 'GKeyFitToScreen':'', 'GKeyShrinkToFit':'', 'GKeyIntervalU':'', 'GKeyIntervalD':'', 'GKeyNamePosition1':'', 'GKeyNamePosition2':'', 'GKeyNamePosition3':'', 'GKeyNamePosition4':'', 'GKeyNamePosition5':'', 'GKeyNamePosition6':'', 'GKeyNamePosition7':'', 'GKeyNamePosition8':'', 'GKeyNamePosition9':'', 'GKeyTextUsesBG':'', 'GKeyPreload':''}
OriginalSettings=dict(Settings) #Clone default settings in case an invalid setting is received and needs to be restored to default
FileList=[] #The "File List" (see PictureViewer.cfg) - This is filled in during ProcessSettings()

#Setting file formatting and other information
WhiteSpaceCharacters=' \r\n\t' #Whitespace characters that are stripped/ignored
SettingFileName=u'PictureViewer.cfg' #The filename of the setting file - Make sure to set this as unicode so all of its found information is also unicode
NameValueDelimiter='=' #The character that separates the setting name from its value
CommentCharacter='#' #The comment character that ends a line, when this symbol is found, it and anything after it on the line is ignored. This can also be used on an otherwise blank line

#Shortcut key information
ShortcutKeys={} #This holds a mapping of virtual keys to their shortcut key
GlobalShortcutKeys={} #This holds a mapping of virtual keys to their global shortcut keys
ShortcutKeyList='Quit,NameToCB,Info,Next,Previous,ZoomIn,ZoomOut,Cycle,TranslucencyU,TranslucencyD,DisplayName,Position1,Position2,Position3,Position4,Position5,Position6,Position7,Position8,Position9,Background,IsOpaque,FitToScreen,ShrinkToFit,IntervalU,IntervalD,NamePosition1,NamePosition2,NamePosition3,NamePosition4,NamePosition5,NamePosition6,NamePosition7,NamePosition8,NamePosition9,TextUsesBG,Preload'.split(',') #The list of shortcut key names
ShortcutControlKeys={'ctrl':1<<8, 'alt':1<<9, 'shift':1<<10, 'win':1<<11} #List of control keys and their bit values when used with virtual keys

#Retrieve settings from the options file, and return the settings dictionary
def LoadSettings():
    SettingsFound=set([]) #Holds names of all found settings

    #Open the settings file
    SettingsFile=[] #Blank out settings file variable in case it can't be opened
    try:
        SettingsFile=open(SettingFileName, 'r')
    except IOError as (errno, strerror):
        sys.stderr.write('Cannot open {0} configuration file: I/O error({1}): {2}\n'.format(SettingFileName, errno, strerror))

    #Parse each setting individually from the settings file
    LineNumber=0
    for Setting in SettingsFile:
        LineNumber=LineNumber+1
        #Bypass blank or comment lines
        Setting=Setting.strip(WhiteSpaceCharacters) #Strip whitespace
        if(Setting=='' or Setting[0]==CommentCharacter):
            continue

        #Split setting into separate sections
        SettingName, SettingDelimiter, SettingValue=Setting.partition(NameValueDelimiter) #Split setting around the equal sign
        if(SettingDelimiter==''): #If the setting is not properly formatted, throw an error and continue to next line
            sys.stderr.write('Invalid setting {0} (line {1}): {2}\n'.format('configuration', LineNumber, Setting))
            continue

        #Determine if setting name is valid
        SettingName=SettingName.strip(WhiteSpaceCharacters) #Strip whitespace
        if(SettingName=='' or not (SettingName in Settings)): #Confirm name is an actual setting name
            sys.stderr.write('Invalid setting {0} (line {1}): {2}\n'.format('name', LineNumber, SettingName))
            continue
        if(SettingName in SettingsFound): #Confirm setting has not already been set
            sys.stderr.write('Setting cannot be set twice, first value used (line {0}): {1}\n'.format(LineNumber, SettingName))
            continue

        #Strip the value of whitespace and comments, and make sure its not blank
        SettingValue=SettingValue.strip(WhiteSpaceCharacters) #Strip whitespace
        if(SettingValue.find('#')!=-1): #If there is a comment, remove it, and strip further found whitespace
            SettingValue=SettingValue[0:SettingValue.find('#')].strip(WhiteSpaceCharacters)

        #Save the setting
        SettingsFound.add(SettingName) #Remember that this setting has been found
        Settings[SettingName]=SettingValue

    return Settings

#Save requested settings back to the config file
def SaveSettings(PictureOn):
    #Settings only need to be saved if requested and if they have changed
    if((not Settings['SaveStartOn']) or Settings['StartOn']==PictureOn):
        return

    #Read the data from the file
    try:
        ConfigFile=open(SettingFileName, 'r+')
        ConfigFileData=ConfigFile.read()
    except IOError as (errno, strerror):
        sys.stderr.write('Error saving to {0} configuration file: I/O error({1}): {2}\n'.format(SettingFileName, errno, strerror))
        return

    #Update the StartOn setting
    ReplaceStartOn=re.compile(r'^(\s*StartOn\s*=\s*)\S*', re.MULTILINE)
    UpdatedString=ReplaceStartOn.sub(r'\g<1>'+str(PictureOn), ConfigFileData) #Attempt to replace the StartOn setting
    if(UpdatedString==ConfigFileData): #If replace was not able to be made, setting is assumed not in the file, and is added
        UpdatedString+="\nStartOn="+str(PictureOn)

    #Write the setting back to the file
    try:
        ConfigFile.seek(0)
        ConfigFile.write(UpdatedString)
        ConfigFile.truncate()
        ConfigFile.close()
    except IOError as (errno, strerror):
        sys.stderr.write('Error saving to {0} configuration file: I/O error({1}): {2}\n'.format(SettingFileName, errno, strerror))
        return

#Error string for ProcessSettings and set back to default
def ProcessSettingsError(Name, TypeInfo, NewValue):
    sys.stderr.write('Setting ({0}) of value ({1}) is not a valid {2}, setting back to default ({3})\n'.format(Name, Settings[Name], TypeInfo, OriginalSettings[Name]))
    Settings[Name]=NewValue

#Determine if all settings are valid and process them, and returns the File List array
def ProcessSettings():
    #Process number settings
    for SettingName in ['StartOn', 'Position', 'Translucency', 'Interval', 'NamePosition']: #Convert number settings to integers
        try:
            Settings[SettingName]=int(Settings[SettingName]) #Convert to integer
        except ValueError: #If not a valid number, process error
            ProcessSettingsError(SettingName, 'number', int(OriginalSettings[SettingName]))
    for SettingName, SettingSpan in {'Position':[1,9], 'NamePosition':[1,9], 'Translucency':[0,100], 'Interval':[0,999999]}.iteritems(): #Make sure number settings are inside their valid range
        if(Settings[SettingName]<SettingSpan[0] or Settings[SettingName]>SettingSpan[1]): #If not, process error
            ProcessSettingsError(SettingName, 'number within the range of ({0}-{1})'.format(SettingSpan[0], SettingSpan[1]), int(OriginalSettings[SettingName]))

    #Process boolean settings (convert to booleans)
    BooleanValues={'yes':True, 'true':True, 'on':True, '1':True, 'no':False, 'false':False, 'off':False, '0':False} #Possible boolean values with their result
    for SettingName in ['SaveStartOn', 'Background', 'IsOpaque', 'FitToScreen', 'ShrinkToFit', 'Cycle', 'DisplayName', 'TextUsesBG', 'ExitOnActivity', 'Preload']:
        try:
            Settings[SettingName]=BooleanValues[Settings[SettingName].lower()] #Convert to boolean
        except KeyError: #If not a valid boolean, process error
            ProcessSettingsError(SettingName, 'boolean', BooleanValues[OriginalSettings[SettingName].lower()])

    #Process color settings (convert to integer from hex)
    for SettingName in ['BGColor', 'FGColor']:
        try:
            Settings[SettingName]=int(Settings[SettingName], 16) #Convert to integer
            if(Settings[SettingName]<0 or Settings[SettingName]>0xFFFFFF): #Confirm integer is within color range
                Settings[SettingName]=hex(Settings[SettingName]) #Convert back to hex for proper error
                raise ValueError
        except ValueError: #If not a valid color, process error
            ProcessSettingsError(SettingName, 'color (0-FFFFFF)', int(OriginalSettings[SettingName], 16))

    #Process Extensions setting by turning it into a regular expression of possible matches
    Settings['Extensions']=re.compile('\.('+'|'.join(set(re.sub(r'[^a-zA-Z0-9;]', '', Settings['Extensions']).split(';')))+')$')

    #Process Exclude Path List by splitting it into 3 separate arrays (files, directories, and directories with recursion)
    ExcludeDirList=set() #Directories to be excluded
    ExcludeDirWildcardList=set() #Directories to be excluded recursively
    ExcludeFileList=set() #Files to be excluded
    for Path in Settings['Exclude'].split(';'): #Process each path in the list
        #Format path
        Path=Path.strip(WhiteSpaceCharacters).replace('/','\\') #Strip whitespace and replace / with \
        WildCard=(len(Path)>2 and Path[-2:]=='\\*') #Check for Whildcard
        if(WildCard): #If a wildcard, remove the wildcard character
            Path=Path[:-1]
        if(Path==''): #Skip empty paths
            continue

        #Process the path
        if(os.path.isfile(Path)): #Check for a file
            ExcludeFileList.add(GetRealFileName(Path))
        elif(os.path.isdir(Path)): #Check for a directory
            if(WildCard): #Check whether to recurse into a directory
                ExcludeDirWildcardList.add(GetRealFileName(Path))
            else:
                if(Path[-1]!='\\'): #Make sure directory ends on a "\" (This does not need to be checked for wildcard because that has to end in a "\")
                    Path+="\\"
                ExcludeDirList.add(GetRealFileName(Path))
        else: #If path does not exist, send error
            sys.stderr.write('Invalid path for "{0}": {1}\n'.format('Exclude', Path))

    #Process Include Path List
    IncludeList=[] #The file list
    for Path in Settings['Include'].split(';'): #Process each path in the list
        #Format path
        Path=unicode(Path.strip(WhiteSpaceCharacters).replace('/','\\')) #Strip whitespace and replace / with \
        WildCard=(len(Path)>2 and Path[-2:]=='\\*') #Check for Whildcard
        if(WildCard): #If a wildcard, remove the wildcard character
            Path=Path[:-1]
        if(Path==''): #Skip empty paths
            continue

        #Process the path
        if(os.path.isfile(Path)): #Check for a file
            IncludeList.append(GetRealFileName(Path))
        elif(os.path.isdir(Path)): #Check for a directory
            if(Path[-1]!='\\'): #Make sure directory ends on a "\" (This does not need to be checked for wildcard because that has to end in a "\")
                Path+='\\'
            IncludeList.extend(ProcessPath(GetRealFileName(Path), WildCard, ExcludeDirList, ExcludeDirWildcardList))
        else: #If path does not exist, send error
            sys.stderr.write('Invalid path for "{0}": {1}\n'.format('Include', Path))

    #Remove excluded files
    for FilePath in ExcludeFileList:
        while FilePath in IncludeList: #Remove excluded files as a while loop in case they are in the include list more than once
            del IncludeList[IncludeList.index(FilePath)]

    #If "StartOn" value is too high, process error
    if(Settings['StartOn']<0 or Settings['StartOn']>=len(IncludeList)):
        ProcessSettingsError('StartOn', 'starting value (there are not enough pictures)', 0)

    #Process sort setting
    if(not Settings['Sort'] in ['Path', 'Name', 'Modified', 'Created', 'Random']): #If not a valid sort, process error
        ProcessSettingsError('Sort', 'sort string, which must be one of the values (Path, Name, Modified, Created, Random)', OriginalSettings['Sort'])
    if(Settings['Sort']=='Random'): #If random, we need to include the library
        import random
    #Determine the function used to get keys for sorting files
    FileValueFunc={
        'Path':unicode.lower,
        'Name':lambda x: os.path.basename(x).lower(),
        'Modified':os.path.getmtime,
        'Created':os.path.getctime,
        'Random':lambda x:random.random()
    }[Settings['Sort']]

    #Sort the file list
    TempFileList=[]
    for Path in IncludeList: #Get keys for each file
        TempFileList.append((Path, FileValueFunc(Path))) #Append to list in a key/value tuplet: (key, value)
    TempFileList.sort(key=SortKey) #Do the sorting by the value in the key/value pair
    for Files in TempFileList: #Extract the keys into the final File List
        FileList.append(Files[0])

    #Process (global and local) shortcut keys
    for ShortcutTypes in [(ShortcutKeys, 'LKey'), (GlobalShortcutKeys, 'GKey')]: #Process each type of shortcut keys
        #Prepare shortcut list information for processing
        KeyList=ShortcutTypes[0]
        KeyPrefix=ShortcutTypes[1]
        if(KeyPrefix=='GKey'): #Global keys get one more type of key
            ShortcutKeyList.append('GetFocus')

        for KeyName in ShortcutKeyList: #Process all shortcut keys in the global or local list
            for KeyValue in Settings[KeyPrefix+KeyName].split(';'): #Process key list for a shortcut
                KeyValue=KeyValue.strip(WhiteSpaceCharacters) #Strip whitespace
                try:
                    MyKey=ConvertKey(KeyValue) #Get the virtual key value for this setting
                    if(MyKey==0): #Bypass non set shortcuts
                        pass
                    elif(MyKey in KeyList): #If key has already been used, bypass this shortcut and issue an error
                        sys.stderr.write('Key "{0}" has already been set to "{1}", and cannot be set for "{2}"\n'.format(KeyValue, KeyPrefix+KeyList[MyKey], KeyPrefix+KeyName))
                    else:
                        KeyList[MyKey]=KeyName #Bind the key to the shortcut
                except KeyError: #If an invalid key string is passed, bypass shortcut and issue an error
                    sys.stderr.write('"{0}" is not a valid shortcut key value for "{1}"\n'.format(KeyValue, KeyPrefix+KeyName))

    return FileList

#Find all the files of valid extension in the requested path, recursing if requested
def ProcessPath(Path, Recurse, ExcludeDirList, ExcludeDirWildcardList):
    #Check for path exclusions
    for ExcludePath in ExcludeDirWildcardList: #Check if the current path is a path that is excluded recursively
        if(ExcludePath==Path[0:len(ExcludePath)]):
            return [] #Return nothing for this path
    KeepFilesInThisDirectory=not (Path in ExcludeDirList) #If this path is in the "ExcludeDirList", only exclude files from this directory (but not subdirectories)
    if((not KeepFilesInThisDirectory) and (not Recurse)): #If we are not keeping files in this directory and not recursing, no need to go on
        return [] #Return nothing for this path

    #Get the files
    CurFileList=[] #Receives the files
    try:
        for FileItem in os.listdir(Path): #Recurse over all items in the directory
            if(os.path.isdir(Path+FileItem)): #If a directory...
                if(Recurse): #If recursing, recurse into the directory
                    CurFileList.extend(ProcessPath(Path+FileItem+'\\', True, ExcludeDirList, ExcludeDirWildcardList)) #Add recursed items to returned path list
            elif(KeepFilesInThisDirectory and Settings['Extensions'].search(FileItem)): #If a file and the extension matches (and keeping files in this directory)
                CurFileList.append(GetRealFileName(Path+FileItem)) #Add item to the returned path list
    except WindowsError: #Bypass permission denied errors
        pass

    return CurFileList

#Get original path name (Have to import functions manually since win32api module doesn't seem to support unicode)
import ctypes
GetFullPathName = ctypes.windll.kernel32.GetFullPathNameW
GetLongPathName = ctypes.windll.kernel32.GetLongPathNameW
GetShortPathName = ctypes.windll.kernel32.GetShortPathNameW

#Get a file's real name in windows. This has to be done by converting it to its short path then its long path (after using GetFullPathName to remove extraneous redirections)
def GetRealFileName(Path):
    #Get FullPath
    Path=unicode(Path)
    Buffer=ctypes.create_unicode_buffer(len(Path)*2)
    RetLen=GetFullPathName(Path, len(Buffer), Buffer, None)
    if(RetLen>len(Buffer)):
        Buffer=ctypes.create_unicode_buffer(RetLen)
        GetFullPathName(Path, RetLen, Buffer, None)
    Path=Buffer.value

    #Get ShortPath then LongPath
    for GetXPathName in [GetShortPathName, GetLongPathName]:
        RetLen=GetXPathName(Path, Buffer, len(Buffer))
        if(RetLen>len(Buffer)):
            Buffer=ctypes.create_unicode_buffer(RetLen)
            GetXPathName(Path, Buffer, RetLen)
        Path=Buffer.value

    return Path

#Sort a (key, value) pair by value
def SortKey((k, v)):
    return v

#Convert a key to a virtual key (with control modifiers)
VirtualKeys={ #Virtual keys pulled from WinUser.h
    'VK_0':0x30,'VK_A':0x41,'VK_NUMPAD0':0x60,'VK_F1':0x70, #List only contain their first item
    'VK_LBUTTON':0x01,'VK_RBUTTON':0x02,'VK_CANCEL':0x03,'VK_MBUTTON':0x04,'VK_XBUTTON1':0x05,'VK_XBUTTON2':0x06,'VK_BACK':0x08,'VK_TAB':0x09,'VK_CLEAR':0x0C,'VK_RETURN':0x0D,'VK_SHIFT':0x10,'VK_CONTROL':0x11,'VK_MENU':0x12,'VK_PAUSE':0x13,'VK_CAPITAL':0x14,'VK_KANA':0x15,'VK_HANGEUL':0x15,'VK_HANGUL':0x15,'VK_JUNJA':0x17,'VK_FINAL':0x18,'VK_HANJA':0x19,'VK_KANJI':0x19,'VK_ESCAPE':0x1B,'VK_CONVERT':0x1C,'VK_NONCONVERT':0x1D,'VK_ACCEPT':0x1E,'VK_MODECHANGE':0x1F,'VK_SPACE':0x20,'VK_PRIOR':0x21,'VK_NEXT':0x22,'VK_END':0x23,'VK_HOME':0x24,'VK_LEFT':0x25,'VK_UP':0x26,'VK_RIGHT':0x27,'VK_DOWN':0x28,'VK_SELECT':0x29,'VK_PRINT':0x2A,'VK_EXECUTE':0x2B,'VK_SNAPSHOT':0x2C,'VK_INSERT':0x2D,'VK_DELETE':0x2E,'VK_HELP':0x2F,'VK_LWIN':0x5B,'VK_RWIN':0x5C,'VK_APPS':0x5D,'VK_SLEEP':0x5F,'VK_MULTIPLY':0x6A,'VK_ADD':0x6B,'VK_SEPARATOR':0x6C,'VK_SUBTRACT':0x6D,'VK_DECIMAL':0x6E,'VK_DIVIDE':0x6F,'VK_NUMLOCK':0x90,'VK_SCROLL':0x91,'VK_OEM_NEC_EQUAL':0x92,'VK_OEM_FJ_JISHO':0x92,'VK_OEM_FJ_MASSHOU':0x93,'VK_OEM_FJ_TOUROKU':0x94,'VK_OEM_FJ_LOYA':0x95,'VK_OEM_FJ_ROYA':0x96,'VK_LSHIFT':0xA0,'VK_RSHIFT':0xA1,'VK_LCONTROL':0xA2,'VK_RCONTROL':0xA3,'VK_LMENU':0xA4,'VK_RMENU':0xA5,'VK_BROWSER_BACK':0xA6,'VK_BROWSER_FORWARD':0xA7,'VK_BROWSER_REFRESH':0xA8,'VK_BROWSER_STOP':0xA9,'VK_BROWSER_SEARCH':0xAA,'VK_BROWSER_FAVORITES':0xAB,'VK_BROWSER_HOME':0xAC,'VK_VOLUME_MUTE':0xAD,'VK_VOLUME_DOWN':0xAE,'VK_VOLUME_UP':0xAF,'VK_MEDIA_NEXT_TRACK':0xB0,'VK_MEDIA_PREV_TRACK':0xB1,'VK_MEDIA_STOP':0xB2,'VK_MEDIA_PLAY_PAUSE':0xB3,'VK_LAUNCH_MAIL':0xB4,'VK_LAUNCH_MEDIA_SELECT':0xB5,'VK_LAUNCH_APP1':0xB6,'VK_LAUNCH_APP2':0xB7,'VK_OEM_1':0xBA,'VK_OEM_PLUS':0xBB,'VK_OEM_COMMA':0xBC,'VK_OEM_MINUS':0xBD,'VK_OEM_PERIOD':0xBE,'VK_OEM_2':0xBF,'VK_OEM_3':0xC0,'VK_OEM_4':0xDB,'VK_OEM_5':0xDC,'VK_OEM_6':0xDD,'VK_OEM_7':0xDE,'VK_OEM_8':0xDF,'VK_OEM_AX':0xE1,'VK_OEM_102':0xE2,'VK_ICO_HELP':0xE3,'VK_ICO_00':0xE4,'VK_PROCESSKEY':0xE5,'VK_ICO_CLEAR':0xE6,'VK_PACKET':0xE7,'VK_OEM_RESET':0xE9,'VK_OEM_JUMP':0xEA,'VK_OEM_PA1':0xEB,'VK_OEM_PA2':0xEC,'VK_OEM_PA3':0xED,'VK_OEM_WSCTRL':0xEE,'VK_OEM_CUSEL':0xEF,'VK_OEM_ATTN':0xF0,'VK_OEM_FINISH':0xF1,'VK_OEM_COPY':0xF2,'VK_OEM_AUTO':0xF3,'VK_OEM_ENLW':0xF4,'VK_OEM_BACKTAB':0xF5,'VK_ATTN':0xF6,'VK_CRSEL':0xF7,'VK_EXSEL':0xF8,'VK_EREOF':0xF9,'VK_PLAY':0xFA,'VK_ZOOM':0xFB,'VK_NONAME':0xFC,'VK_PA1':0xFD,'VK_OEM_CLEAR':0xFE
}
SymbolMaps=["`-=[]\:',./!<", 'VK_OEM_3,VK_OEM_MINUS,VK_OEM_PLUS,VK_OEM_4,VK_OEM_6,VK_OEM_5,VK_OEM_1,VK_OEM_7,VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,VK_OEM_8,VK_OEM_102'.split(',')] #Individual symbol mappings
OtherMappings={ #Other key mappings
    'num+':'VK_ADD', 'num.':'VK_DECIMAL', 'num/':'VK_DIVIDE', 'num*':'VK_MULTIPLY', 'num-':'VK_SUBTRACT', 'numlock':'VK_NUMLOCK', #Number pad
    'esc':'VK_ESCAPE', 'pause':'VK_PAUSE', 'break':'VK_CANCEL', 'printscreen':'VK_SNAPSHOT', 'insert':'VK_INSERT', 'delete':'VK_DELETE', 'pageup':'VK_PRIOR', 'pagedown':'VK_NEXT', 'home':'VK_HOME', 'end':'VK_END', 'backspace':'VK_BACK', 'tab':'VK_TAB', 'capslock':'VK_CAPITAL', 'enter':'VK_RETURN', 'up':'VK_UP', 'down':'VK_DOWN', 'left':'VK_LEFT', 'right':'VK_RIGHT', 'space':'VK_SPACE', 'contextmenu':'VK_APPS', 'scrolllock':'VK_SCROLL', #Other normal keys
    'volumedown':'VK_VOLUME_DOWN', 'volumemute':'VK_VOLUME_MUTE', 'volumeup':'VK_VOLUME_UP', 'media':'VK_LAUNCH_MEDIA_SELECT', 'nexttrack':'VK_MEDIA_NEXT_TRACK', 'previoustrack':'VK_MEDIA_PREV_TRACK', 'play_pause':'VK_MEDIA_PLAY_PAUSE', 'stop':'VK_MEDIA_STOP', 'browserback':'VK_BROWSER_BACK', 'browserfavorites':'vk_browser_favorites', 'browserforward':'VK_BROWSER_FORWARD', 'browserhome':'VK_BROWSER_HOME', 'browserrefresh':'VK_BROWSER_REFRESH', 'browsersearch':'VK_BROWSER_SEARCH', 'browserstop':'VK_BROWSER_STOP', 'app1':'VK_LAUNCH_APP1', 'app2':'VK_LAUNCH_APP2', 'mail':'VK_LAUNCH_MAIL', #Vendor specific
    'abnt_c1':'VK_ABNT_C1', 'abnt_c2':'VK_ABNT_C2', 'attn':'VK_ATTN', 'clear':'VK_CLEAR', 'cr_sel':'VK_CRSEL', 'er_eof':'VK_EREOF', 'execute':'VK_EXECUTE', 'ex_sel':'VK_EXSEL', 'icoclr':'VK_ICO_CLEAR', 'icohlp':'VK_ICO_HELP', 'noname':'VK_NONAME', 'oem_attn':'VK_OEM_ATTN', 'auto':'VK_OEM_AUTO', 'ax':'VK_OEM_AX', 'back_tab':'VK_OEM_BACKTAB', 'oemclr':'VK_OEM_CLEAR', 'copy':'VK_OEM_COPY', 'cu_sel':'VK_OEM_CUSEL', 'enlw':'VK_OEM_ENLW', 'finish':'VK_OEM_FINISH', 'loya':'VK_OEM_FJ_LOYA', 'mashu':'VK_OEM_FJ_MASSHOU', 'roya':'VK_OEM_FJ_ROYA', 'touroku':'VK_OEM_FJ_TOUROKU', 'jump':'VK_OEM_JUMP', 'oempa1':'VK_OEM_PA1', 'oempa2':'VK_OEM_PA2', 'oempa3':'VK_OEM_PA3', 'reset':'VK_OEM_RESET', 'wsctrl':'VK_OEM_WSCTRL', 'pa1':'VK_PA1', 'packet':'VK_PACKET', 'play':'VK_PLAY', 'process':'VK_PROCESSKEY', 'select':'VK_SELECT', 'separator':'VK_SEPARATOR', 'zoom':'VK_ZOOM', 'accept':'VK_ACCEPT', 'convert':'VK_CONVERT', 'final':'VK_FINAL', 'help':'VK_HELP', 'ico00':'VK_ICO_00', 'junja':'VK_JUNJA', 'kana':'VK_KANA', 'kanji':'VK_KANJI', 'mode_change':'VK_MODECHANGE', 'non_convert':'VK_NONCONVERT', 'jisho':'VK_OEM_FJ_JISHO', 'print':'VK_PRINT', 'sleep':'VK_SLEEP', #Others
    'lbutton':'VK_LBUTTON', 'rbutton':'VK_RBUTTON', 'mbutton':'VK_MBUTTON', 'xbutton1':'VK_XBUTTON1', 'xbutton2':'VK_XBUTTON2' #Mouse
}
def ConvertKey(KeyStr): #Do the actual converting
    #Convert to upper case for comparison
    KeyStr=KeyStr.lower()

    if(KeyStr=="null" or KeyStr==""): #Check for no shortcut (cannot be combined with control keys)
        return 0

    #Find control keys
    ControlModifiers=0 #This contains the bitwise modifier for the found control keys
    while(KeyStr.find('+')!=-1): #Keep looping while a + is found (meaning there might be more control modifiers)
        try:
            ControlKeyName=KeyStr[0:KeyStr.index('+')] #Find the control key name
            ControlModifiers|=ShortcutControlKeys[ControlKeyName] #Add control key value to the control modifier
            KeyStr=KeyStr[len(ControlKeyName)+1:] #Remove the found control key from the final string
        except KeyError: #If Control Key is not valid, exit the control key loop
            break

    #Check for lists first to save lookup time
    if(len(KeyStr)==1 and KeyStr[0]>='a' and KeyStr[0]<='z'): #Alphabetic
        return VirtualKeys['VK_A']+ord(KeyStr[0])-ord('a')+ControlModifiers
    if(len(KeyStr)==1 and KeyStr[0]>='0' and KeyStr[0]<='9'): #Numeric
        return VirtualKeys['VK_0']+ord(KeyStr[0])-ord('0')+ControlModifiers
    if(len(KeyStr)==4 and KeyStr[0:3]=='num' and KeyStr[3]>='0' and KeyStr[3]<='9'): #Numerpad Numeric
        return VirtualKeys['VK_NUMPAD0']+ord(KeyStr[3])-ord('0')+ControlModifiers
    try: #Function keys 1-24
        if(len(KeyStr)<=3 and KeyStr[0]=='f' and int(KeyStr[1:])>=1 and int(KeyStr[1:])<=24):
            return VirtualKeys['VK_F1']+int(KeyStr[1:])-1+ControlModifiers
    except ValueError: #If an invalid function key number, do not count as function key
        pass

    #Individual symbol mappings
    if(len(KeyStr)==1 and SymbolMaps[0].find(KeyStr)!=-1):
        return VirtualKeys[SymbolMaps[1][SymbolMaps[0].find(KeyStr)]]+ControlModifiers

    return VirtualKeys[OtherMappings[KeyStr]]+ControlModifiers #Returns any other keys