#!python

#Error that is displayed in a message box and forces the program to exit
def CatastrophicError():
    import ctypes, traceback
    ctypes.windll.user32.MessageBoxA(None, "Unexpected error while loading programing: "+"\n".join(traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])), "Cannot open program", 0)
    sys.exit(0)

#Close process on interrupt signals
def SignalHandler(signum, frame):
    win32gui.PostQuitMessage(0)

#Attempt to load all required modules
try:
    import sys #Print errors (sys.stderr)
    from PIL import Image, ImageWin #Image processing
    import GetSettings #Interact with the configuration file and import settings
    
    #Listen for interrupt signals - This does not work as well in Windows as one would hope :-\
    import signal
    signal.signal(signal.SIGABRT, SignalHandler)
    signal.signal(signal.SIGINT, SignalHandler)
    signal.signal(signal.SIGTERM, SignalHandler)

    #Import windows APIs
    import ctypes, win32gui, win32con, win32api, win32clipboard
    RegisterHotKey=ctypes.windll.user32.RegisterHotKey
    SetTimer=ctypes.windll.user32.SetTimer
    KillTimer=ctypes.windll.user32.KillTimer
    GetTextExtentExPoint=ctypes.windll.gdi32.GetTextExtentExPointW
    GetTextExtentPoint32=ctypes.windll.gdi32.GetTextExtentPoint32W
    TextOut=ctypes.windll.gdi32.TextOutW
    GetTickCount=ctypes.windll.kernel32.GetTickCount
    class SIZE(ctypes.Structure):
        _fields_ = [("cx", ctypes.c_int),
                    ("cy", ctypes.c_int)]
except: #If an error occurs while loading a module
    CatastrophicError()

#Send errors to error file
sys.stderr = open('./PictureViewer.err', 'w')

#Main function
State={'ProgramStart':GetTickCount()} #State settings
def Main():
    #Prepare all the application settings and states
    global Settings, FileList
    Settings=GetSettings.LoadSettings() #Retrieve settings from the options file and move Settings to local context
    FileList=GetSettings.ProcessSettings() #Determine if all settings are valid and process them, and move File List to local context

    #Check for empty file list
    if(len(FileList)==0):
        win32api.MessageBox(None, "No pictures found, exiting program", "Error", win32con.MB_OK)
        return

    #Set state settings
    State['ScreenWidth']=win32api.GetSystemMetrics(win32con.SM_CXSCREEN) #Screen width
    State['ScreenHeight']=win32api.GetSystemMetrics(win32con.SM_CYSCREEN) #Screen height
    State['PictureOn']=Settings['StartOn'] #The current picture the program is on in the list
    State['BackgroundBrush']=win32gui.CreateSolidBrush(Settings['BGColor'])
    State['PositionRect']=(0, 0, State['ScreenWidth'], State['ScreenHeight']) #How the window is currently positioned (as a rectangle)
    State['DisplayInfo']=False #Whether we are currently displaying the help information
    State['Timer']=None #The timer object used to cycle images

    #State settings that are filled in elsewhere
    #hwnd: The handle to the window
    #Width/Height: The dimensions of the picture
    #ScaleWidth/ScaleHeight: The scaled dimensions of the picture
    #Scale: The scale at which to display the picture
    #OriginalImageObject: The original PIL image object
    #ImageObject: The DIB section used to hold the image pixel data
    #DrawRect: Where to draw the image in the window (as a rectangle)

    #Load the icon for the window
    FoundIcon=None
    MyExeName=win32api.GetModuleFileName(win32api.GetModuleHandle(None))
    if(MyExeName.find('PictureViewer')!=-1): #Attempt to load from the executable
        try:
            FoundIcon=win32gui.ExtractIcon(win32api.GetModuleHandle(None), MyExeName, 0)
        except:
            FoundIcon=None
    if(FoundIcon==None): #If not found, attempt to load from an icon file in the current directory
        try:
            FoundIcon=win32gui.LoadImage(None, r"HyruleanProductions.ico", win32con.IMAGE_ICON, win32con.LR_DEFAULTSIZE, win32con.LR_DEFAULTSIZE, win32con.LR_LOADFROMFILE)
        except:
            FoundIcon=None

    #Create the window
    wc = win32gui.WNDCLASS()
    wc.lpszClassName = 'PictureViewer'
    wc.style=win32con.CS_VREDRAW|win32con.CS_HREDRAW
    wc.lpfnWndProc={
        win32con.WM_CLOSE:OnClose, win32con.WM_PAINT:OnPaint, win32con.WM_HOTKEY:OnHotKey, win32con.WM_KEYDOWN:OnKeyDown, win32con.WM_TIMER:OnTimer, #Windows messages
        win32con.WM_MOUSEMOVE:ActivityProcessed, win32con.WM_LBUTTONDOWN:ActivityProcessed, win32con.WM_RBUTTONDOWN:ActivityProcessed, win32con.WM_MBUTTONDOWN:ActivityProcessed #Mouse activity
    }
    wc.hIcon=FoundIcon
    wcAtom=win32gui.RegisterClass(wc)
    State['hwnd']=hwnd=win32gui.CreateWindowEx(win32con.WS_EX_TOPMOST, wcAtom, 'Picture Viewer', win32con.WS_VISIBLE|win32con.WS_POPUP, 0, 0, State['ScreenWidth'], State['ScreenHeight'], 0, 0, 0, None)
    UpdateWindowTranslucency()

    #Display the first available picture
    if(State['PictureOn']==0): #Start one back since "AvanceImage" will get back to the image we want to start on
        State['PictureOn']=len(FileList)-1
    else:
        State['PictureOn']=State['PictureOn']-1
    AdvanceImage(1)

    #Process global shortcut keys
    ControlKeyMappings={GetSettings.ShortcutControlKeys['ctrl']:win32con.MOD_CONTROL, GetSettings.ShortcutControlKeys['alt']:win32con.MOD_ALT, GetSettings.ShortcutControlKeys['shift']:win32con.MOD_SHIFT, GetSettings.ShortcutControlKeys['win']:win32con.MOD_WIN} #Internal control keys to windows control keys
    for GlobalKey in GetSettings.GlobalShortcutKeys: #Attempt to register each requested global shortcut key
        #Enumerate the control keys from the virtual keycode
        VirtualKeyCode=GlobalKey & 0xFF
        FoundControlKeys=0
        for ControlMapping in ControlKeyMappings:
            if(GlobalKey & ControlMapping):
                FoundControlKeys+=ControlKeyMappings[ControlMapping]
        #Register the shortcut key to its index in the GlobalShortcutKeys list
        if(not RegisterHotKey(hwnd, GetSettings.ShortcutKeyList.index(GetSettings.GlobalShortcutKeys[GlobalKey]), FoundControlKeys, VirtualKeyCode)):
           sys.stderr.write('One of the global hot keys bound to GKey"{0}" cannot be set\n'.format(GetSettings.GlobalShortcutKeys[GlobalKey]))

    #Run the program until the window is closed and then cleanup the window
    RetVal=win32gui.PumpMessages()
    win32gui.DestroyWindow(hwnd)
    win32gui.UnregisterClass(wcAtom,None)

    #Save requested settings back to the config file
    GetSettings.SaveSettings(State['PictureOn'])

#Draw current screen state
def DrawScreen(hdc):
    if(not 'DrawRect' in State): #If an image has not yet been loaded, wait for it
        return

    #Draw the background if needed
    if(Settings['Background']):
        win32gui.FillRect(hdc, (0, 0, State['ScreenWidth'], State['ScreenHeight']), State['BackgroundBrush'])

    State['ImageObject'].draw(ImageWin.HDC(hdc), State['DrawRect']) #Draw the image

    if(Settings['DisplayName'] or State['DisplayInfo']): #Display text if requested
        #Information used to determine the positions of text
        ViewableWidth=min(State['PositionRect'][2]-State['PositionRect'][0], State['ScreenWidth']) #The width of the viewable window
        ViewableHeight=min(State['PositionRect'][3]-State['PositionRect'][1], State['ScreenHeight']) #The height of the viewable window
        ViewableStartX=Cond(State['ScaleWidth']>State['ScreenWidth'], -State['PositionRect'][0], 0)
        ViewableStartY=Cond(State['ScaleHeight']>State['ScreenHeight'], -State['PositionRect'][1], 0)

        #Prepare the window for drawing text
        win32gui.SetTextColor(hdc, Settings['FGColor']) #Set the foreground color
        win32gui.SetBkColor(hdc, Settings['BGColor']) #Set the background color
        win32gui.SetBkMode(hdc, Cond(Settings['TextUsesBG'], win32con.OPAQUE, win32con.TRANSPARENT)) #Set the background mode

        if(State['DisplayInfo']): #Display info if requested
            #Compile settings to display
            InfoText=''
            for SettingName in ['Translucency','Interval']:
                InfoText+=SettingName+': '+str(Settings[SettingName])+'\n'
            InfoText+='Scale'+': '+'{0:.3f}'.format(State['Scale']*100)+"%\n"
            for SettingName in ['Cycle','Background','IsOpaque','FitToScreen','ShrinkToFit']:
                InfoText+=SettingName+': '+Cond(Settings[SettingName], 'Yes', 'No')+'\n'

            #Draw the info
            InfoHeight=win32gui.DrawText(hdc, InfoText, len(InfoText), (0, 0, ViewableWidth, ViewableHeight), win32con.DT_CALCRECT|win32con.DT_WORDBREAK)[0] #Determine height of info text, length of string is 1 less than needed to remove final line break
            TheHeight=win32gui.DrawText(hdc, InfoText, len(InfoText), (ViewableStartX, ViewableStartY+(ViewableHeight-InfoHeight)/2, ViewableStartX+ViewableWidth, ViewableStartY+ViewableHeight), win32con.DT_CENTER|win32con.DT_WORDBREAK) #Display the info text

        if(Settings['DisplayName']): #Display file name if requested
            #Split the string into multiple lines so they all fit on the screen
            FinalStrings=[] #The list of final strings
            TheString=State['FileName']
            while(TheString!=''):
                MaxChars=ctypes.c_int(0)
                Size=SIZE()
                GetTextExtentExPoint(hdc, TheString.encode('utf-16'), len(TheString)+1, ViewableWidth, ctypes.pointer(MaxChars), None, ctypes.pointer(Size))
                if(MaxChars.value<2): #If no more characters can be displayed on the line because the picture is too small, exit prematurely
                    break
                FinalStrings.append(TheString[0:MaxChars.value-1])
                TheString=TheString[MaxChars.value-1:]

            #Calculate the lengths of each of the strings
            FinalStringsSizes=[]
            TotalStringsHeight=0
            for TheString in FinalStrings:
                Size=SIZE()
                GetTextExtentPoint32(hdc, TheString, len(TheString)+1, ctypes.pointer(Size))
                FinalStringsSizes.append((Size.cx, Size.cy))
                TotalStringsHeight+=Size.cy

            #Draw each of the strings
            PositionAxisX=(Settings['NamePosition']-1)%3 #Where the text gets positioned horizontally (0=left, 1=center, 2=right)
            PositionAxisY=(Settings['NamePosition']-1)/3 #Where the text gets positioned vertically (0=bottom, 1=center, 2=top)
            StringDrawTop=ViewableStartY+[ViewableHeight-TotalStringsHeight, (ViewableHeight-TotalStringsHeight)/2, 0][PositionAxisY] #Calculate the top of where the strings are going to be displayed
            for i in range(0,len(FinalStrings)):
                TheString=FinalStrings[i]
                StringSize=FinalStringsSizes[i]
                StringDrawLeft=ViewableStartX+[0, (ViewableWidth-StringSize[0])/2, ViewableWidth-StringSize[0]][PositionAxisX]
                TextOut(hdc, StringDrawLeft, StringDrawTop, TheString.encode('utf-16'), len(TheString)+1)
                StringDrawTop+=StringSize[1]

#Load the picture to display and draw the image
def LoadPicture():
    if('DrawRect' in State): #Do not display image until it is ready
        del State['DrawRect']

    State['FileName']=FileList[State['PictureOn']] #Get the picture name

    #Load the picture
    try:
        State['OriginalImage']=TheImage=Image.open(State['FileName']).convert("RGB") #Load the picture from a file
        (State['Width'], State['Height'])=TheImage.size #Save the image dimensions
        State['ImageObject']=ImageWin.Dib(TheImage) #Turn the image into a windows DIB section for quick drawing (Do it this first time to make sure the image is valid)
    except:
        sys.stderr.write('Could not load image: {0}\n'.format(State['FileName'].encode('utf-8')))
        raise NameError('LoadPictureException')

    #Figure out how the image is to be displayed on the screen and draw the image
    DetermineImageScaling()

#Determine the scale to use on the image based on settings (and call to reposition and draw the image)
def DetermineImageScaling():
    State['Scale']=1.0 #Start the image scale at 100%
    if(Settings['FitToScreen'] or (Settings['ShrinkToFit'] and (State['Width']>State['ScreenWidth'] or State['Height']>State['ScreenHeight']))): #Check if the picture needs to be scaled to fit the screen
        ScreenAspectRatio=float(State['ScreenWidth'])/State['ScreenHeight'] #Determine the screen's aspect ratio
        PictureAspectRatio=float(State['Width'])/State['Height'] #Determine the pictures aspect ratio
        if(PictureAspectRatio>ScreenAspectRatio): #If the picture's aspect ratio is larger than the screen, then the width is the constraining member
            State['Scale']=float(State['ScreenWidth'])/State['Width']
        else: #Otherwise the height is the constraining member
            State['Scale']=float(State['ScreenHeight'])/State['Height']

    DetermineImageSize()

#Determine the image's scaled dimensions and turn the image into a windows DIB section for quick drawing (and call to reposition and draw the image)
def DetermineImageSize():
    State['ScaleWidth']=int(State['Width']*State['Scale'])
    State['ScaleHeight']=int(State['Height']*State['Scale'])
    if(State['Scale']==1.0): #Only resize the stored image if need be
        State['ImageObject']=ImageWin.Dib(State['OriginalImage'])
    else: #Resize the image if it is not a 1.0 scale
        State['ImageObject']=ImageWin.Dib(State['OriginalImage'].resize((State['ScaleWidth'], State['ScaleHeight']), Cond(State['Scale']>1.0, Image.BICUBIC, Image.ANTIALIAS)))

    DetermineImagePosition()

#Figure out how the image is to be displayed on the screen (and call to draw the image)
def DetermineImagePosition():
    #Set State variables locally
    (ScaleWidth, ScaleHeight, ScreenWidth, ScreenHeight)=(State['ScaleWidth'], State['ScaleHeight'], State['ScreenWidth'], State['ScreenHeight'])

    #Determine where to place the picture on the screen (final rectangle)
    PositionAxisX=(Settings['Position']-1)%3 #0=left, 1=center, 2=right
    PositionAxisY=(Settings['Position']-1)/3 #0=bottom, 1=center, 2=top
    PositionRect=( #Determine the final rectangle by the 3 positions each of the x/y axis can be on
        [0, (ScreenWidth-ScaleWidth)/2, ScreenWidth-ScaleWidth][PositionAxisX], #Left
        [ScreenHeight-ScaleHeight, (ScreenHeight-ScaleHeight)/2, 0][PositionAxisY], #Top
        [ScaleWidth, (ScreenWidth+ScaleWidth)/2, ScreenWidth][PositionAxisX], #Right
        [ScreenHeight, (ScreenHeight+ScaleHeight)/2, ScaleHeight][PositionAxisY] #Bottom
    )

    #The window and drawing position are determined by if the background is showing
    if(Settings['Background']):
        State['DrawRect']=PositionRect
        PositionRect=(0, 0, ScreenWidth, ScreenHeight) #Full screen
    else:
        State['DrawRect']=(0, 0, ScaleWidth, ScaleHeight)

    #Reposition the window if needed
    if(PositionRect!=State['PositionRect']):
        State['PositionRect']=PositionRect
        win32gui.SetWindowPos(State['hwnd'], 0, PositionRect[0], PositionRect[1], PositionRect[2]-PositionRect[0], PositionRect[3]-PositionRect[1], win32con.SWP_NOACTIVATE) #Make sure window stays on top

    RedrawScreen() #Forces a redraw of the screen

def UpdateWindowTranslucency():
    hwnd=State['hwnd']
    if(not Settings['IsOpaque']): #Set window as translucent
        win32api.SetWindowLong(hwnd, win32con.GWL_EXSTYLE, win32api.GetWindowLong(hwnd, win32con.GWL_EXSTYLE) | win32con.WS_EX_TRANSPARENT+win32con.WS_EX_LAYERED) #Set window to translucent mode
        win32gui.SetLayeredWindowAttributes(hwnd, 0, int(Settings['Translucency']/float(100)*255), win32con.LWA_ALPHA) #Set translucency value
    else: #Set window as opaque
        win32api.SetWindowLong(hwnd, win32con.GWL_EXSTYLE, win32api.GetWindowLong(hwnd, win32con.GWL_EXSTYLE) & ~(win32con.WS_EX_TRANSPARENT+win32con.WS_EX_LAYERED)) #Set window to translucent mode

    win32gui.SetWindowPos(hwnd, win32con.HWND_TOPMOST, 0, 0, 0, 0, win32con.SWP_SHOWWINDOW|win32con.SWP_NOACTIVATE|win32con.SWP_NOMOVE|win32con.SWP_NOSIZE) #Make sure window stays on top

def AdvanceImage(Direction):
    StartPictureOn=State['PictureOn'] #Remember the first picture
    while(True): #Loop until a picture can be displayed
        State['PictureOn']+=Direction #Advance the picture number
        if(Direction>0): #Make sure picture number didn't overflow list boundaries
            if(State['PictureOn']>=len(FileList)):
                State['PictureOn']=0
        else:
            if(State['PictureOn']<0):
                State['PictureOn']=len(FileList)-1

        try: #Attempt to load the picture
           LoadPicture()
           return
        except NameError as (TheError):
            if(TheError.args[0]!='LoadPictureException'): #Confirm error type
                raise TheError

            if(StartPictureOn==State['PictureOn']): #If picture loops around to start picture, no pictures can be displayed, so exit program
                win32api.MessageBox(State['hwnd'], "All pictures were invalid, exiting program", "Error", win32con.MB_OK)
                win32gui.PostQuitMessage(0)
                return

#Forces a redraw of the screen
def RedrawScreen():
    win32gui.InvalidateRect(State['hwnd'], None, False)
    ResetTimer() #After redrawing an image, reset the timer

#Start up a new timer (reset old one to make sure enough time passes after an event has happened)
def ResetTimer():
    if(State['Timer']!=None): #If a timer is running cancel it
        KillTimer(State['hwnd'], State['Timer'])
    if(Settings['Cycle']): #If cycling, start a timer
        State['Timer']=SetTimer(State['hwnd'], 1, Settings['Interval'], None)
    else: #Otherwise set the timer to unused
        State['Timer']=None

#Window messages
def OnClose(hwnd, msg, wp, lp): #Window closing
    win32gui.PostQuitMessage(0)
def OnPaint(hwnd, msg, wp, lp): #Window paint requested
    dc, ps=win32gui.BeginPaint(hwnd)
    DrawScreen(dc)
    win32gui.EndPaint(hwnd, ps)
    return 0
def OnHotKey(hwnd, msg, wp, lp): #Global (Hot) key was pressed
    ProcessShortCut(GetSettings.ShortcutKeyList[wp])
def OnTimer(hwnd, msg, wp, lp): #Global (Hot) key was pressed
    AdvanceImage(1)
def ActivityProcessed(hwnd, msg, wp, lp): #Key or mouse was used
    #Exit the program if set to do so on activity
    if(Settings['ExitOnActivity'] and (msg!=win32con.WM_MOUSEMOVE or GetTickCount()-State['ProgramStart']>2000)): #Must wait at least 2 seconds before this message is triggered from a mouse move (was triggering prematurely)
        win32gui.PostQuitMessage(0)
def OnKeyDown(hwnd, msg, wp, lp): #Local key was pressed
    ActivityProcessed(0,0,0,0)

    #Get Modifiers
    ControlModifiers=0 #This contains the bitwise modifier for the found control keys
    CheckKeys={win32con.VK_CONTROL:'ctrl', win32con.VK_MENU:'alt', win32con.VK_SHIFT:'shift', win32con.VK_LWIN:'win', win32con.VK_RWIN:'win'}
    for CheckKey in CheckKeys: #Test each control key
        if(win32api.GetAsyncKeyState(CheckKey)&0x80000000): #If control key is down
            ControlModifiers|=GetSettings.ShortcutControlKeys[CheckKeys[CheckKey]]
    VirtualKey=wp #Get the virtual key code

    #Check to see if the key is mapped to a shortcut
    FinalKey=VirtualKey|ControlModifiers
    if(FinalKey in GetSettings.ShortcutKeys):
        ProcessShortCut(GetSettings.ShortcutKeys[FinalKey]) #If so, process it

#Process keys
def ProcessShortCut(KeyName):
    #Handle non boolean settings first
    if(KeyName=='Quit'):
        win32gui.PostQuitMessage(0)
    elif(KeyName=='GetFocus'):
        win32gui.SetForegroundWindow(State['hwnd'])
    elif(KeyName=='IntervalU' or KeyName=='IntervalD'):
        Settings['Interval']=max(Settings['Interval']+Cond(KeyName=='IntervalU', 100, -100), 100) #Go up or down by 100ms on the interval (with the lowest value of 100)
        if(State['DisplayInfo']):
            RedrawScreen()
        elif(Settings['Cycle']):
            ResetTimer()
    elif(KeyName=='TranslucencyU' or KeyName=='TranslucencyD'):
        Settings['Translucency']=min(max(Settings['Translucency']+Cond(KeyName=='TranslucencyU', 1, -1), 0),100) #Increase or decrease translucency by 1% (clamp between 0 and 100)
        if(not Settings['IsOpaque']):
            UpdateWindowTranslucency()
        if(State['DisplayInfo']):
            RedrawScreen()
    elif(KeyName=='ZoomIn' or KeyName=='ZoomOut'):
        State['Scale']=min(max(State['Scale']*Cond(KeyName=='ZoomIn', 1.1, .9), .005),10000) #Scale up or down by 10% percent (clamp between .005 and 10000)
        DetermineImageSize()
    elif(KeyName=='Next'):
        AdvanceImage(1)
    elif(KeyName=='Previous'):
        AdvanceImage(-1)
    elif(KeyName=='NameToCB'):
        win32clipboard.OpenClipboard()
        win32clipboard.SetClipboardData(win32con.CF_UNICODETEXT, State['FileName'])
        win32clipboard.CloseClipboard()
    elif(KeyName[0:8]=='Position'):
        Settings['Position']=int(KeyName[8])
        DetermineImagePosition()
    elif(KeyName[0:12]=='NamePosition'):
        Settings['NamePosition']=int(KeyName[12])
        if(Settings['DisplayName']):
            RedrawScreen()
    elif(KeyName=='Info'):
        State['DisplayInfo']=not State['DisplayInfo']
        RedrawScreen()
    else: #Boolean settings
        Settings[KeyName]=not Settings[KeyName] #All boolean settings swap their value
        if(KeyName=='DisplayName'):
            RedrawScreen()
        elif(KeyName=='Cycle'):
            if(State['DisplayInfo']):
                RedrawScreen()
            else:
                ResetTimer()
        elif(KeyName=='Background'):
            DetermineImagePosition()
        elif(KeyName=='IsOpaque'):
            UpdateWindowTranslucency()
            if(State['DisplayInfo']):
                RedrawScreen()
        elif(KeyName=='FitToScreen' or KeyName=='ShrinkToFit'):
            DetermineImageScaling()
        elif(KeyName=='TextUsesBG'):
            if(Settings['DisplayName'] or State['DisplayInfo']):
                RedrawScreen()
        elif(KeyName=='Preload'):
            if(State['DisplayInfo']):
                RedrawScreen()
        else:
            raise ("Invalid ShortcutKey Message", KeyName)

def Cond(IsTrue, IfTrue, IfFalse): #Conditional
    if(IsTrue):
        return IfTrue
    return IfFalse

#Run the program
try:
    Main()
except:
    CatastrophicError()